@@ -1,7 +1,7 @@
 /*
  * image viewer, for framebuffer devices
  *
- *   (c) 1998-2004 Gerd Knorr <kraxel@bytesex.org>
+ *   (c) 1998-2004 Gerd Hoffmann <kraxel@bytesex.org>
  *
  */
 
@@ -69,6 +69,7 @@
 /* with arg */
 #define KEY_GOTO     -100
 #define KEY_SCALE    -101
+#define KEY_DELAY    -102
 
 /* edit */
 #define KEY_DELETE   -200
@@ -84,10 +85,16 @@ int lirc = -1;
 
 /* variables for read_image */
 int32_t         lut_red[256], lut_green[256], lut_blue[256];
-int             dither = FALSE, pcd_res = 3, steps = 50;
+int             dither = FALSE, pcd_res = 3;
+int             v_steps = 50;
+int             h_steps = 50;
+int             text_steps;
 int             textreading = 0, redraw = 0, statusline = 1;
 int             new_image;
 int             left, top;
+int             show_top;
+int             show_bottom;
+int             fitwidth;
 
 /* file list */
 struct flist {
@@ -113,6 +120,7 @@ static float fbgamma = 1;
 /* Command line options. */
 int autodown;
 int autoup;
+int autofirst;
 int comments;
 int transparency = 40;
 
@@ -127,7 +135,7 @@ version(void)
 {
     fprintf(stderr,
 	    "fbi version " VERSION ", compiled on %s\n"
-	    "(c) 1999-2004 Gerd Knorr <kraxel@bytesex.org> [SUSE Labs]\n",
+	    "(c) 1999-2004 Gerd Hoffmann <kraxel@bytesex.org> [SUSE Labs]\n",
 	    __DATE__ );
 }
 
@@ -148,8 +156,8 @@ usage(char *name)
 	    "\n",
 	    name);
 
-    cfg_help_cmdline(fbi_cmd,4,20,0);
-    cfg_help_cmdline(fbi_cfg,4,20,40);
+    cfg_help_cmdline(stderr,fbi_cmd,4,20,0);
+    cfg_help_cmdline(stderr,fbi_cfg,4,20,40);
 
     fprintf(stderr,
 	    "\n"
@@ -230,6 +238,11 @@ static struct flist* flist_first(void)
     return list_entry(flist.next, struct flist, list);
 }
 
+static struct flist* flist_last(void)
+{
+    return list_entry(flist.prev, struct flist, list);
+}
+
 static struct flist* flist_next(struct flist *f, int eof, int loop)
 {
     if (flist_islast(f)) {
@@ -242,10 +255,13 @@ static struct flist* flist_next(struct flist *f, int eof, int loop)
     return list_entry(f->list.next, struct flist, list);
 }
 
-static struct flist* flist_prev(struct flist *f)
+static struct flist* flist_prev(struct flist *f, int loop)
 {
-    if (flist_isfirst(f))
+    if (flist_isfirst(f)) {
+	if (loop)
+	    return flist_last();
 	return f;
+    }
     return list_entry(f->list.prev, struct flist, list);
 }
 
@@ -536,9 +552,10 @@ static void debug_key(char *key)
 
     len = sprintf(linebuffer,"key: ");
     for (i = 0; key[i] != '\0'; i++)
-	len += sprintf(linebuffer+len, "%s%c",
-		       key[i] < 0x20 ? "^" : "",
-		       key[i] < 0x20 ? key[i] + 0x40 : key[i]);
+	len += snprintf(linebuffer+len, sizeof(linebuffer)-len,
+			"%s%c",
+			key[i] < 0x20 ? "^" : "",
+			key[i] < 0x20 ? key[i] + 0x40 : key[i]);
     status_update(NULL, linebuffer, NULL);
 }
 
@@ -612,7 +629,8 @@ read_image(char *filename)
     }
     if (NULL == loader) {
 	/* no loader found, try to use ImageMagick's convert */
-	sprintf(command,"convert -depth 8 \"%s\" ppm:-",filename);
+	snprintf(command,sizeof(command),
+		 "convert -depth 8 \"%s\" ppm:-",filename);
 	if (NULL == (fp = popen(command,"r")))
 	    return NULL;
 	loader = &ppm_loader;
@@ -677,6 +695,8 @@ static float auto_scale(struct ida_image *img)
     float xs,ys,scale;
     
     xs = (float)fb_var.xres / img->i.width;
+    if (fitwidth)
+	return xs;
     ys = (float)fb_var.yres / img->i.height;
     scale = (xs < ys) ? xs : ys;
     return scale;
@@ -687,7 +707,7 @@ static float auto_scale(struct ida_image *img)
 static int
 svga_show(struct ida_image *img, int timeout, char *desc, char *info, int *nr)
 {
-    static int      paused = 0;
+    static int      paused = 0, skip = KEY_SPACE;
     int             exif = 0, help = 0;
     int             rc;
     char            key[11];
@@ -698,17 +718,22 @@ svga_show(struct ida_image *img, int timeout, char *desc, char *info, int *nr)
 
     *nr = 0;
     if (NULL == img)
-	return KEY_SPACE; /* skip */
+	return skip;
     
     if (new_image) {
 	/* start with centered image, if larger than screen */
 	if (img->i.width > fb_var.xres)
 	    left = (img->i.width - fb_var.xres) / 2;
 	if (img->i.height > fb_var.yres) {
-	    if (textreading)
+	    top = (img->i.height - fb_var.yres) / 2;
+	    if (show_top) {
+		show_top = 0;
 		top = 0;
-	    else
-		top = (img->i.height - fb_var.yres) / 2;
+	    }
+	    if (show_bottom) {
+		show_bottom = 0;
+		top = img->i.height - fb_var.yres;
+	    }
 	}
 	new_image = 0;
     }
@@ -794,17 +819,18 @@ svga_show(struct ida_image *img, int timeout, char *desc, char *info, int *nr)
 	} else if (0 == strcmp(key, " ")) {
 	    if (textreading && top < (int)(img->i.height - fb_var.yres)) {
 		redraw = 1;
-		top += (fb_var.yres-100);
+		top += text_steps;
 	    } else {
+		skip = KEY_SPACE;
 		return KEY_SPACE;
 	    }
 
 	} else if (0 == strcmp(key, "\x1b[A") && img->i.height > fb_var.yres) {
 	    redraw = 1;
-	    top -= steps;
+	    top -= v_steps;
 	} else if (0 == strcmp(key, "\x1b[B") && img->i.height > fb_var.yres) {
 	    redraw = 1;
-	    top += steps;
+	    top += v_steps;
 	} else if (0 == strcmp(key, "\x1b[1~") && img->i.height > fb_var.yres) {
 	    redraw = 1;
 	    top = 0;
@@ -813,17 +839,34 @@ svga_show(struct ida_image *img, int timeout, char *desc, char *info, int *nr)
 	    top = img->i.height - fb_var.yres;
 	} else if (0 == strcmp(key, "\x1b[D") && img->i.width > fb_var.xres) {
 	    redraw = 1;
-	    left -= steps;
+	    left -= h_steps;
 	} else if (0 == strcmp(key, "\x1b[C") && img->i.width > fb_var.xres) {
 	    redraw = 1;
-	    left += steps;
+	    left += h_steps;
+
+	} else if (0 == strcmp(key, "\x1b[5~") ||
+		   0 == strcmp(key, "j")       ||
+		   0 == strcmp(key, "J")) {
+	    if (textreading && top > 0) {
+		redraw = 1;
+		top -= text_steps;
+	    } else {
+		skip = KEY_PGUP;
+		return KEY_PGUP;
+	    }
 
-	} else if (0 == strcmp(key, "\x1b[5~")) {
-	    return KEY_PGUP;
 	} else if (0 == strcmp(key, "\x1b[6~") ||
+		   0 == strcmp(key, "k")       ||
+		   0 == strcmp(key, "K")       ||
 		   0 == strcmp(key, "n")       ||
 		   0 == strcmp(key, "N")) {
-	    return KEY_PGDN;
+	    if (textreading && top < (int)(img->i.height - fb_var.yres)) {
+		redraw = 1;
+		top += text_steps;
+	    } else {
+		skip = KEY_PGDN;
+		return KEY_PGDN;
+	    }
 	    
 	} else if (0 == strcmp(key, "+")) {
 	    return KEY_PLUS;
@@ -883,9 +926,11 @@ svga_show(struct ida_image *img, int timeout, char *desc, char *info, int *nr)
 	    return KEY_GOTO;
 	} else if (rc == 1 && (*key == 's' || *key == 'S')) {
 	    return KEY_SCALE;
+	} else if (rc == 1 && (*key == 'x' || *key == 'X')) {
+	    return KEY_DELAY;
 	} else if (rc == 1 && *key >= '0' && *key <= '9') {
 	    *nr = *nr * 10 + (*key - '0');
-	    sprintf(linebuffer, "> %d",*nr);
+	    snprintf(linebuffer, sizeof(linebuffer), "> %d",*nr);
 	    status_update(img, linebuffer, NULL);
 	} else {
 	    *nr = 0;
@@ -1080,7 +1125,7 @@ static void edit_desc(struct ida_image *img, char *filename)
 static void cleanup_and_exit(int code)
 {
     shadow_fini();
-    fb_clear_mem();
+    fb_clear_screen();
     tty_restore();
     fb_cleanup();
     flist_print_tagged(stdout);
@@ -1140,13 +1185,16 @@ main(int argc, char *argv[])
     once        = GET_ONCE();
     autoup      = GET_AUTO_UP();
     autodown    = GET_AUTO_DOWN();
+    autofirst   = GET_AUTO_FIRST();
+    fitwidth    = GET_FIT_WIDTH();
     statusline  = GET_VERBOSE();
     textreading = GET_TEXT_MODE();
     editable    = GET_EDIT();
     backup      = GET_BACKUP();
     preserve    = GET_PRESERVE();
 
-    steps       = GET_SCROLL();
+    v_steps     = GET_SCROLL();
+    h_steps     = GET_SCROLL();
     timeout     = GET_TIMEOUT();
     pcd_res     = GET_PCD_RES();
 
@@ -1188,10 +1236,16 @@ main(int argc, char *argv[])
     shadow_set_palette(fd);
     signal(SIGTSTP,SIG_IGN);
     
+    if (textreading) {
+	show_top = 1;
+	text_steps = fb_var.yres - 100;
+    }
+
     /* svga main loop */
     tty_raw();
     desc = NULL;
     info = NULL;
+    scale = 1;
     for (;;) {
 	if (need_read) {
 	    need_read = 0;
@@ -1200,10 +1254,9 @@ main(int argc, char *argv[])
 	    fimg = NULL;
 	    simg = NULL;
 	    img  = NULL;
-	    sprintf(linebuffer,"loading %s ...",fcurrent->name);
+	    snprintf(linebuffer,sizeof(linebuffer),"loading %s ...",fcurrent->name);
 	    status_update(img,linebuffer, NULL);
 	    fimg = read_image(fcurrent->name);
-	    scale = 1;
 	    if (fimg) {
 		if (autoup || autodown) {
 		    scale = auto_scale(fimg);
@@ -1211,10 +1264,13 @@ main(int argc, char *argv[])
 			scale = 1;
 		    if (scale > 1 && !autoup)
 			scale = 1;
+		    if (autofirst)
+			autoup = autodown = 0;
 		}
 		if (scale != 1) {
-		    sprintf(linebuffer,"scaling (%.0f%%) %s ...",
-			    scale*100, fcurrent->name);
+		    snprintf(linebuffer, sizeof(linebuffer),
+			     "scaling (%.0f%%) %s ...",
+			     scale*100, fcurrent->name);
 		    status_update(img,linebuffer, NULL);
 		    simg = scale_image(fimg,scale);
 		    img = simg;
@@ -1224,20 +1280,24 @@ main(int argc, char *argv[])
 		desc = make_desc(&fimg->i,fcurrent->name);
 	    }
 	    if (!img) {
-		sprintf(linebuffer,"%s: FAILED",fcurrent->name);
+		snprintf(linebuffer,sizeof(linebuffer),
+			 "%s: FAILED",fcurrent->name);
 		status_error(linebuffer);
 	    }
 	}
-	if (img)
+	if (img) {
+	    int pages = ceil((float)img->i.height / fb_var.yres);
+	    text_steps = ceil((float)img->i.height / pages);
 	    info = make_info(fimg,scale);
+	}
 	switch (key = svga_show(img, timeout, desc, info, &arg)) {
 	case KEY_DELETE:
 	    if (editable) {
 		struct flist *fdel = fcurrent;
 		if (flist_islast(fcurrent))
-			fcurrent = flist_prev(fcurrent);
+		    fcurrent = flist_prev(fcurrent,0);
 		else
-			fcurrent = flist_next(fcurrent,0,0);
+		    fcurrent = flist_next(fcurrent,0,0);
 		unlink(fdel->name);
 		flist_del(fdel);
 		flist_renumber();
@@ -1254,7 +1314,8 @@ main(int argc, char *argv[])
 	case KEY_ROT_CCW:
 	{
 	    if (editable) {
-		sprintf(linebuffer,"rotating %s ...",fcurrent->name);
+		snprintf(linebuffer,sizeof(linebuffer),
+			 "rotating %s ...",fcurrent->name);
 		status_update(img, linebuffer, NULL);
 		jpeg_transform_inplace
 		    (fcurrent->name,
@@ -1278,6 +1339,8 @@ main(int argc, char *argv[])
 	case KEY_SPACE:
 	    need_read = 1;
 	    fcurrent = flist_next(fcurrent,1,0);
+	    if (textreading)
+		show_top = 1;
 	    if (NULL != fcurrent)
 		break;
 	    /* else fall */
@@ -1288,11 +1351,15 @@ main(int argc, char *argv[])
 	    break;
 	case KEY_PGDN:
 	    need_read = 1;
-	    fcurrent = flist_next(fcurrent,0,0);
+	    fcurrent = flist_next(fcurrent,0,1);
+	    if (textreading)
+		show_top = 1;
 	    break;
 	case KEY_PGUP:
 	    need_read = 1;
-	    fcurrent = flist_prev(fcurrent);
+	    fcurrent = flist_prev(fcurrent,1);
+	    if (textreading)
+		show_bottom = 1;
 	    break;
 	case KEY_TIMEOUT:
 	    need_read = 1;
@@ -1313,7 +1380,7 @@ main(int argc, char *argv[])
 	    } else if (key == KEY_ASCALE) {
 		newscale = auto_scale(fimg);
 	    } else {
-		newscale = arg / 100;
+		newscale = arg / 100.0;
 	    }
 	    if (newscale < 0.1)
 		newscale = 0.1;
@@ -1321,8 +1388,9 @@ main(int argc, char *argv[])
 		newscale = 10;
 	    scale_fix_top_left(scale, newscale, img);
 	    scale = newscale;
-	    sprintf(linebuffer,"scaling (%.0f%%) %s ...",
-		    scale*100, fcurrent->name);
+	    snprintf(linebuffer,sizeof(linebuffer),
+		     "scaling (%.0f%%) %s ...",
+		     scale*100, fcurrent->name);
 	    status_update(NULL, linebuffer, NULL);
 	    free_image(simg);
 	    simg = scale_image(fimg,scale);
@@ -1334,6 +1402,9 @@ main(int argc, char *argv[])
 		fcurrent = flist_goto(arg);
 	    }
 	    break;
+	case KEY_DELAY:
+	    timeout = arg;
+	    break;
 	case KEY_VERBOSE:
 #if 0 /* fbdev testing/debugging hack */
 	    {
