@@ -60,6 +60,16 @@ struct cfg_cmdline fbi_cfg[] = {
 	.option   = { O_AUTO_DOWN },
 	.yesno    = 1,
 	.desc     = "  like the above, but downscale only",
+    },{
+	.cmdline  = "autofirst",
+	.option   = { O_AUTO_FIRST },
+	.yesno    = 1,
+	.desc     = "  autoscale first image only",
+    },{
+	.cmdline  = "fitwidth",
+	.option   = { O_FIT_WIDTH },
+	.yesno    = 1,
+	.desc     = "  use width only for autoscaling",
 
     },{
 	.letter   = 'v',
